import io
import json
from contextlib import contextmanager
from os import PathLike
from pathlib import Path
from typing import IO, Any, Dict, Generator, Protocol  # noqa: I101

from ._optional_libs import hcl2, toml, yaml


StrPathIO = str | PathLike | IO | io.IOBase


class ConfigLoadCallable(Protocol):  # pylint: disable=too-few-public-methods
    def __call__(self, f: StrPathIO, **kw: Any) -> Any:
        pass  # pragma: no cover


class ConfigLoadsCallable(Protocol):  # pylint: disable=too-few-public-methods
    def __call__(self, s: str, **kw: Any) -> Any:
        pass  # pragma: no cover


class ConfigDumpCallable(Protocol):  # pylint: disable=too-few-public-methods
    def __call__(self, obj: Dict[str, Any], f: StrPathIO, **kw: Any) -> None:
        pass  # pragma: no cover


class ConfigDumpsCallable(Protocol):  # pylint: disable=too-few-public-methods
    def __call__(self, obj: Dict[str, Any], **kw: Any) -> str:
        pass  # pragma: no cover


@contextmanager
def ensure_stream(
    stream: StrPathIO, write: bool = False
) -> Generator[IO | io.IOBase, None, None]:
    if isinstance(stream, io.IOBase):
        yield stream
    else:
        f = io.open(
            Path(stream).expanduser(),  # type: ignore[arg-type]
            (write and 'w' or 'r') + 't',
            encoding='utf-8',
            errors='surrogateescape',
        )
        yield f
        f.close()


# === JSON ===


def json_dump(obj: Dict[str, Any], f: StrPathIO, **kw: Any) -> None:
    ensure_ascii = kw.pop('ensure_ascii', False)
    with ensure_stream(f, write=True) as s:
        json.dump(
            obj, s, ensure_ascii=ensure_ascii, **kw  # type: ignore[arg-type]
        )


def json_dumps(obj: Dict[str, Any], **kw: Any) -> str:
    sio = io.StringIO()
    json_dump(obj, sio, **kw)
    return sio.getvalue()


def json_load(f: StrPathIO, **kw: Any) -> Any:
    with ensure_stream(f) as s:
        return json.load(s, **kw)


def json_loads(s: str, **kw: Any) -> Any:
    return json.loads(s, **kw)


# === YAML ===


def yaml_dump(obj: Dict[str, Any], f: StrPathIO, **kw: Any) -> None:
    allow_unicode = kw.pop('allow_unicode', True)
    encoding = kw.pop('encoding', 'utf-8')

    with ensure_stream(f, write=True) as s:
        yaml.dump(
            obj,
            s,
            yaml.SafeDumper,
            allow_unicode=allow_unicode,
            encoding=encoding,
            **kw,
        )


def yaml_dumps(obj: Dict[str, Any], **kw: Any) -> str:
    sio = io.StringIO()
    yaml_dump(obj, sio, **kw)
    return sio.getvalue()


def yaml_load(f: StrPathIO, **_: Any) -> Any:
    with ensure_stream(f) as s:
        return yaml.load(s, yaml.SafeLoader)


def yaml_loads(s: str, **_: Any) -> Any:
    return yaml.load(s, yaml.SafeLoader)


# === TOML ===


def toml_dump(obj: Dict[str, Any], f: StrPathIO, **kw: Any) -> None:
    with ensure_stream(f, write=True) as s:
        toml.dump(obj, s, **kw)


def toml_dumps(obj: Dict[str, Any], **kw: Any) -> str:
    return toml.dumps(obj, **kw)


def toml_load(f: StrPathIO, **kw: Any) -> Any:
    with ensure_stream(f) as s:
        return toml.load(s, **kw)


def toml_loads(s: str, **kw: Any) -> Any:
    return toml.loads(s, **kw)


# === HCL2 ===


def hcl2_load(f: StrPathIO, **_: Any) -> Any:
    with ensure_stream(f) as s:
        return hcl2.load(s)


def hcl2_loads(s: str, **_: Any) -> Any:
    return hcl2.loads(s)



from __future__ import annotations

from collections.abc import MutableMapping
from typing import Any, AnyStr, Iterator, Optional


class LowerCaseDict(MutableMapping[AnyStr, Any]):
    __slots__ = ('_data',)

    _data: MutableMapping[AnyStr, Any]

    def __init__(
        self, data: Optional[MutableMapping[AnyStr, Any]] = None, **kw: Any
    ) -> None:
        self._data = {}
        self.update(data or {}, **kw)

    def __setitem__(self, key: AnyStr, value: Any) -> None:
        self._data[key.lower()] = value

    def __getitem__(self, key: AnyStr) -> Any:
        return self._data[key.lower()]

    def __delitem__(self, key: AnyStr) -> None:
        del self._data[key.lower()]

    def __iter__(self) -> Iterator[Any]:
        return iter(self._data)

    def __len__(self) -> int:
        return len(self._data)

    def __repr__(self) -> str:
        return f"{type(self).__name__}({dict(self.items())})"



# pylint: disable=no-name-in-module

from __future__ import annotations

import os
from abc import ABCMeta, abstractmethod
from os import PathLike
from pathlib import Path
from types import ModuleType
from typing import (  # noqa: I101
    TYPE_CHECKING,
    Any,
    ClassVar,
    Dict,
    Optional,
    Protocol,
    Sequence,
    Type,
    no_type_check,
)
from warnings import warn

from pydantic import BaseModel
from pydantic.env_settings import InitSettingsSource
from pydantic.fields import SHAPE_SINGLETON, ModelField  # noqa: I101

from ._optional_libs import dotenv  # noqa: I202
from ._optional_libs import hcl2, toml, yaml
from .dumploads import (
    ConfigLoadCallable,
    hcl2_load,
    json_load,
    toml_load,
    yaml_load,
)


if TYPE_CHECKING:
    from ipl_config import BaseSettings  # pragma: no cover


class SettingsStrategyCallable(
    Protocol
):  # pylint: disable=too-few-public-methods
    def __call__(
        self, clazz: Type[BaseSettings] | BaseSettings
    ) -> Dict[str, Any]:
        pass  # pragma: no cover


class SettingsStrategyMetaclass(ABCMeta):  # noqa: B024
    @no_type_check
    def __call__(cls, *args, **kwargs):
        """
        Ensure the class has all its dependencies during constructor
        without needs to call to `super`
        """
        for dep in getattr(cls, '__dependencies__', None) or ():
            if isinstance(dep, Exception):
                raise dep
        return super().__call__(*args, **kwargs)


# type: ignore[too-few-public-methods]
# pylint: disable=too-few-public-methods
class SettingsStrategy(metaclass=SettingsStrategyMetaclass):
    __dependencies__: ClassVar[
        Sequence[ModuleType | ImportError] | None
    ] = None


# pylint: disable=too-few-public-methods
class EnvSettingsStrategy(SettingsStrategy):
    __slots__ = 'env_prefix', 'env_vars', 'case_sensitive'

    def __init__(
        self,
        env_prefix: Optional[str] = None,
        env_vars: Dict[str, Optional[str]] | None = None,
        case_sensitive: Optional[bool] = False,
    ):
        self.env_prefix: Optional[str] = env_prefix
        self.env_vars: Dict[str, Optional[str]] = (
            env_vars  # type: ignore[assignment]
            if env_vars is not None
            else os.environ
        )
        self.case_sensitive: Optional[bool] = case_sensitive

        # self.env_vars: Dict[str, Optional[str]] = {
        #     **os.environ,
        #     **(read_env_file(
        #         env_file, encoding=env_file_encoding
        #     ) if env_file else {}),
        # }

        if not case_sensitive:
            if env_prefix:
                self.env_prefix = env_prefix.lower()
            self.env_vars = {k.lower(): v for k, v in self.env_vars.items()}

    def __call__(
        self, clazz: Type[BaseSettings] | BaseSettings
    ) -> Dict[str, Any]:
        return self._from_env(clazz, self.env_prefix)

    def _from_env(
        self, clazz: Type[BaseModel] | BaseModel, prefix: Optional[str] = None
    ) -> Dict[str, Any]:
        prefix = prefix or ''
        result: Dict[str, Any] = {}

        field: ModelField
        for field in clazz.__fields__.values():
            extra = field.field_info.extra
            if extra.get('deprecated'):
                warn(f"{field.name!r} is deprecated", DeprecationWarning)

            env_name = extra.get('env')
            if not env_name:
                env_name = prefix + (prefix and '_' or '') + field.name
            if field.shape == SHAPE_SINGLETON and issubclass(
                field.type_, BaseModel
            ):
                env_val = self._from_env(field.type_, prefix=env_name)
            else:
                env_val = self.env_vars.get(  # type: ignore[assignment]
                    env_name if self.case_sensitive else env_name.lower()
                )

            if env_val is not None:
                result[field.alias] = env_val

        return result


# pylint: disable=too-few-public-methods
class DotEnvSettingsStrategy(EnvSettingsStrategy):
    __slots__ = ()

    def __init__(
        self,
        env_prefix: Optional[str] = None,
        case_sensitive: Optional[bool] = False,
        env_file: Optional[str | PathLike] = None,
        env_file_encoding: Optional[str] = None,
    ):
        super().__init__(
            env_prefix=env_prefix,
            case_sensitive=case_sensitive,
            env_vars=read_env_file(env_file, encoding=env_file_encoding)
            if env_file
            else {},
        )


# pylint: disable=too-few-public-methods
class KwSettingsStrategy(SettingsStrategy, InitSettingsSource):
    def __init__(self, **kw: Any) -> None:
        super().__init__(init_kwargs=kw)


class FileSettingsStrategy(SettingsStrategy):
    __slots__ = 'path', 'config_format'

    __extensions__: ClassVar[Sequence[str]] = ()

    def __init__(
        self, path: str | PathLike, config_format: Optional[str] = None
    ):
        self.path: Path = Path(path).expanduser()
        self.config_format: Optional[str] = config_format

    def __call__(
        self, clazz: Type[BaseSettings] | BaseSettings
    ) -> Dict[str, Any]:
        if not self.is_acceptable(self.path, self.config_format):
            return {}

        loader = self.get_loader(clazz)
        return loader(self.path)

    @abstractmethod
    def get_loader(
        self, clazz: Type[BaseSettings] | BaseSettings
    ) -> ConfigLoadCallable:
        pass  # pragma: no cover

    @classmethod
    def is_acceptable(
        cls, path: Optional[Path] = None, config_format: Optional[str] = None
    ) -> bool:
        """
        :return: True if format is acceptable to loads
        """
        if path is None:
            return False

        if config_format is None:
            config_format = cls.detect_format(path)

        return bool(config_format and config_format in cls.__extensions__)

    @classmethod
    def detect_format(cls, path: Path) -> Optional[str]:
        return path.suffix[1:]


class JsonSettingsStrategy(FileSettingsStrategy):
    __extensions__ = 'json', 'js'

    def get_loader(
        self, clazz: Type[BaseSettings] | BaseSettings
    ) -> ConfigLoadCallable:
        return json_load


class YamlSettingsStrategy(FileSettingsStrategy):
    __dependencies__ = (yaml,)
    __extensions__ = 'yaml', 'yml'

    def get_loader(
        self, clazz: Type[BaseSettings] | BaseSettings
    ) -> ConfigLoadCallable:
        return yaml_load


class TomlSettingsStrategy(FileSettingsStrategy):
    __dependencies__ = (toml,)
    __extensions__ = 'toml', 'tml'

    def get_loader(
        self, clazz: Type[BaseSettings] | BaseSettings
    ) -> ConfigLoadCallable:
        return toml_load


class Hcl2SettingsStrategy(FileSettingsStrategy):
    __dependencies__ = (hcl2,)
    __extensions__ = 'hcl', 'hcl2', 'tf'

    def get_loader(
        self, clazz: Type[BaseSettings] | BaseSettings
    ) -> ConfigLoadCallable:
        return hcl2_load


def read_env_file(
    path: str | PathLike, *, encoding: Optional[str] = None
) -> Dict[str, Optional[str]]:
    path = Path(path).expanduser()
    is_env_default = str(path) == '.env'
    is_env_exists = path.is_file()

    if is_env_exists and isinstance(dotenv, Exception):
        warn(str(dotenv), ImportWarning)
    elif not is_env_exists and not is_env_default:
        warn(f"{str(path)!r} is not a file", UserWarning)
    else:
        return dotenv.dotenv_values(path, encoding=encoding or 'utf-8')

    return {}



# pylint: disable=no-name-in-module

import sys
from collections import OrderedDict
from decimal import Decimal
from os import PathLike
from pathlib import Path
from typing import Any, ClassVar, Dict, Optional, Sequence, Type

from pydantic import BaseConfig, BaseModel
from pydantic.config import Extra
from pydantic.utils import deep_update

from .dumploads import StrPathIO, json_dump
from .source import (
    DotEnvSettingsStrategy,
    EnvSettingsStrategy,
    Hcl2SettingsStrategy,
    JsonSettingsStrategy,
    KwSettingsStrategy,
    SettingsStrategyCallable,
    TomlSettingsStrategy,
    YamlSettingsStrategy,
)


class BaseSettings(BaseModel):
    __slots__ = ()

    class Config(BaseConfig):  # pylint: disable=too-few-public-methods
        env_prefix: Optional[str] = 'APP'
        env_file: Optional[str | PathLike] = '.env'
        env_file_encoding: Optional[str] = None
        case_sensitive: bool = False
        validate_all: bool = True
        extra: Extra = Extra.ignore
        arbitrary_types_allowed = True

    __config__: ClassVar[Type[Config]] = Config

    def __init__(  # pylint: disable=too-many-arguments
        self,
        env_prefix: Optional[str] = None,
        env_file: Optional[str | PathLike] = None,
        config_file: Optional[str | PathLike] = None,
        config_format: Optional[str] = None,
        source_strategies: Sequence[SettingsStrategyCallable] | None = None,
        **kw: Any,
    ) -> None:
        if config_file is not None:
            config_file = Path(config_file)

        if source_strategies is None:
            cfg = self.__config__
            source_strategies = [
                KwSettingsStrategy(**kw),
                EnvSettingsStrategy(
                    env_prefix=env_prefix or cfg.env_prefix,
                    case_sensitive=cfg.case_sensitive,
                ),
                DotEnvSettingsStrategy(
                    env_prefix=env_prefix or cfg.env_prefix,
                    env_file=env_file or cfg.env_file,
                    env_file_encoding=cfg.env_file_encoding,
                    case_sensitive=cfg.case_sensitive,
                ),
            ]
            if config_file:
                for s in (
                    JsonSettingsStrategy,
                    YamlSettingsStrategy,
                    TomlSettingsStrategy,
                    Hcl2SettingsStrategy,
                ):
                    if s.is_acceptable(config_file, config_format):
                        source_strategies.append(
                            s(
                                config_file,
                                config_format,  # type: ignore[abstract]
                            )
                        )
                        break
                else:
                    raise NotImplementedError(
                        f"No readers found for the config file: {config_file}"
                    )

        super().__init__(
            **deep_update(*reversed([s(self) for s in source_strategies]))
        )

    def _write_json(self, o: Dict[str, Any], f: StrPathIO, **kw: Any) -> None:
        # take encoder from `self.__config__.json_encoders`
        encoder = kw.pop('default', self.__json_encoder__)
        json_dump(o, f, default=encoder, **kw)

    def write_schema(self, f: StrPathIO = sys.stdout, **kw: Any) -> None:
        return self._write_json(self.schema(), f, **kw)

    def write_json(self, f: StrPathIO = sys.stdout, **kw: Any) -> None:
        return self._write_json(self.dict(), f, **kw)

    def to_env(self, **kw: Any) -> Dict[str, str]:
        prefix = self.__config__.env_prefix
        if not prefix:
            return self._to_env(self.dict(**kw))
        return self._to_env(self.dict(**kw), prefix)

    def _to_env(self, ns: Dict[str, Any], *path: str) -> Dict[str, str]:
        res = OrderedDict()

        for name, model in ns.items():
            _path = path + (name,)
            env_name = '_'.join(_.upper() for _ in _path)

            if not isinstance(model, dict):
                # for mypy
                env_val: bool | bytes | str | int | float | complex | Decimal
                if model is None:
                    env_val = ''
                elif isinstance(model, bool):
                    env_val = int(model)
                elif isinstance(model, bytes):
                    env_val = model.decode(
                        encoding='utf-8', errors='surrogateescape'
                    )
                elif isinstance(model, (str, int, float, complex, Decimal)):
                    env_val = model
                else:
                    raise NotImplementedError(
                        f"Not a scalar: {env_name}={type(model)}"
                    )

                res[env_name] = str(env_val)
            else:
                for k, v in self._to_env(model, *_path).items():
                    res[k] = v

        return res



try:
    import dotenv  # type: ignore[import]
except ImportError:
    dotenv = ImportError(  # type: ignore[no-redef]
        'python-dotenv is not installed'
    )

try:
    import hcl2  # type: ignore[import]
except ImportError:
    hcl2 = ImportError(  # type: ignore[no-redef]
        'python-hcl2 is not installed'
    )

try:
    import toml  # type: ignore[import]
except ImportError:
    toml = ImportError('toml is not installed')  # type: ignore[no-redef]

try:
    import yaml  # type: ignore[import]
except ImportError:
    yaml = ImportError('pyyaml is not installed')  # type: ignore[no-redef]


__all__ = 'dotenv', 'hcl2', 'toml', 'yaml'



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ipl_config']

package_data = \
{'': ['*']}

install_requires = \
['pydantic>=1.10.2,<2.0.0']

extras_require = \
{'dotenv': ['python-dotenv>=0.21.0,<0.22.0'],
 'hcl2': ['python-hcl2>=3.0.5,<4.0.0'],
 'toml': ['toml>=0.10.2,<0.11.0'],
 'yaml': ['pyyaml>=6.0,<7.0']}

setup_kwargs = {
    'name': 'ipl-config',
    'version': '0.1.3',
    'description': 'InPlat config adapters',
    'long_description': '[![tests](https://github.com/koi8-r/ipl-config/actions/workflows/ci.yml/badge.svg)](https://github.com/koi8-r/ipl-config/actions/workflows/ci.yml)\n[![codecov](https://codecov.io/gh/koi8-r/ipl-config/branch/master/graph/badge.svg?token=OKURU75Y7A)](https://codecov.io/gh/koi8-r/ipl-config)\n[![pypi](https://img.shields.io/pypi/v/ipl-config.svg)](https://pypi.python.org/pypi/ipl-config)\n[![versions](https://img.shields.io/pypi/pyversions/ipl-config.svg)](https://github.com/koi8-r/ipl-config)\n\n\n# Config adapters with pydantic behavior\n- json\n- yaml\n- toml\n- hcl2\n- environ\n- .env\n\n## Examples\n- [yaml config with env, dotenv and args overrides](/examples/config_yaml_dump.py)\n',
    'author': 'InPlat',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/koi8-r/ipl-config',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
